# Generated from ImpLanguage.g4 by ANTLR 4.11.1
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .ImpLanguageParser import ImpLanguageParser
else:
    from ImpLanguageParser import ImpLanguageParser

# This class defines a complete listener for a parse tree produced by ImpLanguageParser.
class ImpLanguageListener(ParseTreeListener):

    # Enter a parse tree produced by ImpLanguageParser#start.
    def enterStart(self, ctx:ImpLanguageParser.StartContext):
        pass

    # Exit a parse tree produced by ImpLanguageParser#start.
    def exitStart(self, ctx:ImpLanguageParser.StartContext):
        pass


    # Enter a parse tree produced by ImpLanguageParser#add.
    def enterAdd(self, ctx:ImpLanguageParser.AddContext):
        pass

    # Exit a parse tree produced by ImpLanguageParser#add.
    def exitAdd(self, ctx:ImpLanguageParser.AddContext):
        pass


    # Enter a parse tree produced by ImpLanguageParser#sub.
    def enterSub(self, ctx:ImpLanguageParser.SubContext):
        pass

    # Exit a parse tree produced by ImpLanguageParser#sub.
    def exitSub(self, ctx:ImpLanguageParser.SubContext):
        pass


    # Enter a parse tree produced by ImpLanguageParser#cst.
    def enterCst(self, ctx:ImpLanguageParser.CstContext):
        pass

    # Exit a parse tree produced by ImpLanguageParser#cst.
    def exitCst(self, ctx:ImpLanguageParser.CstContext):
        pass


    # Enter a parse tree produced by ImpLanguageParser#var.
    def enterVar(self, ctx:ImpLanguageParser.VarContext):
        pass

    # Exit a parse tree produced by ImpLanguageParser#var.
    def exitVar(self, ctx:ImpLanguageParser.VarContext):
        pass


    # Enter a parse tree produced by ImpLanguageParser#apar.
    def enterApar(self, ctx:ImpLanguageParser.AparContext):
        pass

    # Exit a parse tree produced by ImpLanguageParser#apar.
    def exitApar(self, ctx:ImpLanguageParser.AparContext):
        pass


    # Enter a parse tree produced by ImpLanguageParser#bge.
    def enterBge(self, ctx:ImpLanguageParser.BgeContext):
        pass

    # Exit a parse tree produced by ImpLanguageParser#bge.
    def exitBge(self, ctx:ImpLanguageParser.BgeContext):
        pass


    # Enter a parse tree produced by ImpLanguageParser#bpar.
    def enterBpar(self, ctx:ImpLanguageParser.BparContext):
        pass

    # Exit a parse tree produced by ImpLanguageParser#bpar.
    def exitBpar(self, ctx:ImpLanguageParser.BparContext):
        pass


    # Enter a parse tree produced by ImpLanguageParser#bor.
    def enterBor(self, ctx:ImpLanguageParser.BorContext):
        pass

    # Exit a parse tree produced by ImpLanguageParser#bor.
    def exitBor(self, ctx:ImpLanguageParser.BorContext):
        pass


    # Enter a parse tree produced by ImpLanguageParser#ble.
    def enterBle(self, ctx:ImpLanguageParser.BleContext):
        pass

    # Exit a parse tree produced by ImpLanguageParser#ble.
    def exitBle(self, ctx:ImpLanguageParser.BleContext):
        pass


    # Enter a parse tree produced by ImpLanguageParser#blt.
    def enterBlt(self, ctx:ImpLanguageParser.BltContext):
        pass

    # Exit a parse tree produced by ImpLanguageParser#blt.
    def exitBlt(self, ctx:ImpLanguageParser.BltContext):
        pass


    # Enter a parse tree produced by ImpLanguageParser#band.
    def enterBand(self, ctx:ImpLanguageParser.BandContext):
        pass

    # Exit a parse tree produced by ImpLanguageParser#band.
    def exitBand(self, ctx:ImpLanguageParser.BandContext):
        pass


    # Enter a parse tree produced by ImpLanguageParser#bcst.
    def enterBcst(self, ctx:ImpLanguageParser.BcstContext):
        pass

    # Exit a parse tree produced by ImpLanguageParser#bcst.
    def exitBcst(self, ctx:ImpLanguageParser.BcstContext):
        pass


    # Enter a parse tree produced by ImpLanguageParser#bgt.
    def enterBgt(self, ctx:ImpLanguageParser.BgtContext):
        pass

    # Exit a parse tree produced by ImpLanguageParser#bgt.
    def exitBgt(self, ctx:ImpLanguageParser.BgtContext):
        pass


    # Enter a parse tree produced by ImpLanguageParser#beq.
    def enterBeq(self, ctx:ImpLanguageParser.BeqContext):
        pass

    # Exit a parse tree produced by ImpLanguageParser#beq.
    def exitBeq(self, ctx:ImpLanguageParser.BeqContext):
        pass


    # Enter a parse tree produced by ImpLanguageParser#bnot.
    def enterBnot(self, ctx:ImpLanguageParser.BnotContext):
        pass

    # Exit a parse tree produced by ImpLanguageParser#bnot.
    def exitBnot(self, ctx:ImpLanguageParser.BnotContext):
        pass


    # Enter a parse tree produced by ImpLanguageParser#aff.
    def enterAff(self, ctx:ImpLanguageParser.AffContext):
        pass

    # Exit a parse tree produced by ImpLanguageParser#aff.
    def exitAff(self, ctx:ImpLanguageParser.AffContext):
        pass


    # Enter a parse tree produced by ImpLanguageParser#err.
    def enterErr(self, ctx:ImpLanguageParser.ErrContext):
        pass

    # Exit a parse tree produced by ImpLanguageParser#err.
    def exitErr(self, ctx:ImpLanguageParser.ErrContext):
        pass


    # Enter a parse tree produced by ImpLanguageParser#assert.
    def enterAssert(self, ctx:ImpLanguageParser.AssertContext):
        pass

    # Exit a parse tree produced by ImpLanguageParser#assert.
    def exitAssert(self, ctx:ImpLanguageParser.AssertContext):
        pass


    # Enter a parse tree produced by ImpLanguageParser#loop.
    def enterLoop(self, ctx:ImpLanguageParser.LoopContext):
        pass

    # Exit a parse tree produced by ImpLanguageParser#loop.
    def exitLoop(self, ctx:ImpLanguageParser.LoopContext):
        pass


    # Enter a parse tree produced by ImpLanguageParser#skip.
    def enterSkip(self, ctx:ImpLanguageParser.SkipContext):
        pass

    # Exit a parse tree produced by ImpLanguageParser#skip.
    def exitSkip(self, ctx:ImpLanguageParser.SkipContext):
        pass


    # Enter a parse tree produced by ImpLanguageParser#ite.
    def enterIte(self, ctx:ImpLanguageParser.IteContext):
        pass

    # Exit a parse tree produced by ImpLanguageParser#ite.
    def exitIte(self, ctx:ImpLanguageParser.IteContext):
        pass


    # Enter a parse tree produced by ImpLanguageParser#macrocall.
    def enterMacrocall(self, ctx:ImpLanguageParser.MacrocallContext):
        pass

    # Exit a parse tree produced by ImpLanguageParser#macrocall.
    def exitMacrocall(self, ctx:ImpLanguageParser.MacrocallContext):
        pass


    # Enter a parse tree produced by ImpLanguageParser#seq.
    def enterSeq(self, ctx:ImpLanguageParser.SeqContext):
        pass

    # Exit a parse tree produced by ImpLanguageParser#seq.
    def exitSeq(self, ctx:ImpLanguageParser.SeqContext):
        pass


    # Enter a parse tree produced by ImpLanguageParser#macrodef.
    def enterMacrodef(self, ctx:ImpLanguageParser.MacrodefContext):
        pass

    # Exit a parse tree produced by ImpLanguageParser#macrodef.
    def exitMacrodef(self, ctx:ImpLanguageParser.MacrodefContext):
        pass



del ImpLanguageParser


import functools
from typing import Optional, cast, TypeVar, Callable

from wise.imp import *
from wise.symex import SymStore, Id, NonEmptySymStore

T = TypeVar("T")


def eassert(expr: T, assertion: Callable[[T], bool]) -> T:
    assert assertion(expr)
    return expr


def is_unsat(expr: Bexpr) -> bool:
    solver = z3.Solver()
    solver.add(expr.to_smt())
    return solver.check() == z3.unsat


S = TypeVar("S", Aexpr, Bexpr)


def simplify_expr(expr: S) -> S:
    result = convert_z3_expr(z3.simplify(expr.to_smt()))
    return expr if result is None else result


def simplify_store(store: SymStore) -> SymStore:
    if isinstance(store, Id):
        return store
    assert isinstance(store, NonEmptySymStore)
    return NonEmptySymStore(
        store.var, simplify_expr(store.val), simplify_store(store.s)
    )


def convert_z3_expr(expr: z3.ExprRef) -> Optional[Aexpr | Bexpr]:
    if z3.is_int_value(expr):
        return Cst(cast(z3.IntVal, expr).as_long())
    elif z3.is_const(expr):
        return Var(str(expr))

    children_results = list(map(convert_z3_expr, expr.children()))
    if any(child_result is None for child_result in children_results):
        return None

    match expr.decl().kind():
        case z3.Z3_OP_TRUE:
            return Bcst(True)
        case z3.Z3_OP_FALSE:
            return Bcst(False)
        case z3.Z3_OP_LE:
            return Ble(*children_results)
        case z3.Z3_OP_LT:
            return Band(Ble(*children_results), Bnot(Beq(*children_results)))
        case z3.Z3_OP_GE:
            return Ble(*reversed(list(children_results)))
        case z3.Z3_OP_GT:
            return Band(
                Ble(*reversed(list(children_results))), Bnot(Beq(*children_results))
            )
        case z3.Z3_OP_EQ:
            return Beq(*children_results)
        case z3.Z3_OP_NOT:
            return Bnot(*children_results)
        case z3.Z3_OP_AND:
            return functools.reduce(
                lambda acc, elem: cast(Bexpr, Band(elem, acc)),
                reversed(children_results),
            )
        case z3.Z3_OP_OR:
            return functools.reduce(
                lambda acc, elem: cast(Bexpr, Bor(elem, acc)),
                reversed(children_results),
            )
        case z3.Z3_OP_ADD:
            return Add(*children_results)
        case z3.Z3_OP_SUB:
            return Sub(*children_results)
        case _:
            return None



"""
(** A symbolic state is a path constraint expressed as a [bexpr] and a symbolic store *)
Definition sym_state := (bexpr * sym_store)%type.
"""
from abc import abstractmethod
from dataclasses import dataclass
from typing import Tuple

from wise.imp import Bexpr, Ident, Aexpr, Var, Cst, Add, Sub, Bcst, Ble, Beq, Band, \
    Bnot, Imp


@dataclass(frozen=True)
class SymStore:
    """
    (** Symbolic stores are simply substitutions (of variable with expressions) *)
    Definition sym_store :=
      ident -> aexpr.
    """

    @abstractmethod
    def __call__(self, x: Ident) -> Aexpr:
        raise NotImplementedError()


@dataclass(frozen=True)
class Id(SymStore):
    """
    Definition id : sym_store := fun x => Var x.
    """

    def __call__(self, x: Ident) -> Aexpr:
        return Var(x)

    def __str__(self):
        return "{}"


@dataclass(frozen=True)
class NonEmptySymStore(SymStore):
    var: Ident
    val: Aexpr
    s: SymStore = Id()

    def __call__(self, x: Ident) -> Aexpr:
        return self.val if x == self.var else self.s(x)

    def __str__(self):
        return f"{self.s}{{{self.var} -> {self.val}}}"


# (** A symbolic state is a path constraint expressed as a [bexpr],
#     a symbolic store and a program to execute
# *)
# Definition sym_state := (bexpr * sym_store * IMP)%type.
SymState = Tuple[Bexpr, SymStore, Imp]


def sym_update(s: SymStore, x: Ident, e: Aexpr) -> SymStore:
    """
    (** Update a symbolic store *)
    Definition sym_update (s : sym_store) (x : ident) (e : aexpr) : sym_store :=
      fun y => if string_dec y x then e else s y.
    """
    return NonEmptySymStore(x, e, s)


def sym_aeval(s: SymStore, e: Aexpr) -> Aexpr:
    """
    (** Symbolic evaluation of arithmetic expressions *)
    Fixpoint sym_aeval (s : sym_store) (e : aexpr) : aexpr :=
      match e with
      | Var x => s x
      | Cst c => e
      | Add e1 e2 => Add (sym_aeval s e1) (sym_aeval s e2)
      | Sub e1 e2 => Sub (sym_aeval s e1) (sym_aeval s e2)
      end.
    """

    match e:
        case Var(x):
            return s(x)
        case Cst(_):
            return e
        case Add(e1, e2):
            return Add(sym_aeval(s, e1), sym_aeval(s, e2))
        case Sub(e1, e2):
            return Sub(sym_aeval(s, e1), sym_aeval(s, e2))


def sym_beval(s: SymStore, e: Bexpr) -> Bexpr:
    """
    (** Symbolic evaluation of boolean expressions *)
    Fixpoint sym_beval (s : sym_store) (e : bexpr) : bexpr :=
      match e with
      | Bcst b      => e
      | Ble e1 e2   => Ble (sym_aeval s e1) (sym_aeval s e2)
      | Beq e1 e2   => Beq (sym_aeval s e1) (sym_aeval s e2)
      | Band e1 e2  => Band (sym_beval s e1) (sym_beval s e2)
      | Bnot e      => Bnot (sym_beval s e)
      end.
    """
    match e:
        case Bcst(_):
            return e
        case Ble(e1, e2):
            return Ble(sym_aeval(s, e1), sym_aeval(s, e2))
        case Beq(e1, e2):
            return Beq(sym_aeval(s, e1), sym_aeval(s, e2))
        case Band(e1, e2):
            return Band(sym_beval(s, e1), sym_beval(s, e2))
        case Bnot(e):
            return Bnot(sym_beval(s, e))



import unittest
from typing import List, cast

from wise.bugfinder import (
    find_bugs,
    BugFound,
    find_bugs_depth_first,
    find_bugs_assuming,
    Finished,
)
from wise.helpers import is_unsat, simplify_expr, simplify_store
from wise.imp import *
from wise.parser import parse_imp, parse_bexpr
from wise.streams import peek
from wise.symex import NonEmptySymStore, Id


class TestBugfinder(unittest.TestCase):
    def test_simple_bug(self):
        """
        Example my_prog :=
          Ite
            (Beq (Var "x"%string) (Cst 0%Z))
            Err
            Skip.

        Compute (peek 10 (find_bugs my_prog)).
        """
        my_prog = Ite(Beq(Var("x"), Cst(0)), Err(), Skip())

        bug_path = Band(
            left=Bcst(b=True),
            right=Beq(left=Var(ident="x"), right=Cst(num=0)),
        )
        bug_store = Id()
        bug = BugFound((bug_path, bug_store, Err()))

        self.assertIn(bug, peek(10, find_bugs(my_prog)))

    def test_simple_bug_2(self):
        my_prog = Ite(Beq(Var("x"), Cst(0)), Seq(Aff("x", Cst(17)), Err()), Skip())

        bug_path = Band(
            left=Bcst(b=True),
            right=Beq(left=Var(ident="x"), right=Cst(num=0)),
        )
        bug_store = NonEmptySymStore(var="x", val=Cst(num=17), s=Id())
        bug = BugFound((bug_path, bug_store, Err()))

        self.assertIn(bug, peek(10, find_bugs(my_prog)))

    def test_gcd_correct(self):
        with open("../examples/gcd_correct.imp", "r") as f:
            gcd_correct = parse_imp(f.read())

        assumptions = parse_bexpr("a > 0 and b > 0")
        result = peek(100, find_bugs_assuming(gcd_correct, assumptions))

        self.assertTrue(not any(isinstance(elem, BugFound) for elem in result))

    def test_gcd_buggy(self):
        with open("../examples/gcd_buggy.imp", "r") as f:
            gcd_buggy = parse_imp(f.read())

        assumptions = parse_bexpr("a > 0 and b > 0")
        result = peek(60, find_bugs_assuming(gcd_buggy, assumptions))
        bugs: List[BugFound] = cast(
            List[BugFound], list(filter(BugFound.__instancecheck__, result))
        )

        # 0 <= a and not (0 == a) and 0 <= b and not (0 == b) and not (a <= 0) and
        #   not (b <= 0) and not (a == b) and not (a <= b) and not (0 <= (a + b)
        #   and 0 <= b and b <= 0 and not (b == 0))
        simplified_bug_path = Band(
            Ble(Cst(0), Var("a")),
            Band(
                Bnot(Beq(Cst(0), Var("a"))),
                Band(
                    Ble(Cst(0), Var("b")),
                    Band(
                        Bnot(Beq(Cst(0), Var("b"))),
                        Band(
                            Bnot(Ble(Var("a"), Cst(0))),
                            Band(
                                Bnot(Ble(Var("b"), Cst(0))),
                                Band(
                                    Bnot(Beq(Var("a"), Var("b"))),
                                    Band(
                                        Bnot(Ble(Var("a"), Var("b"))),
                                        Bnot(
                                            Band(
                                                Ble(Cst(0), Add(Var("a"), Var("b"))),
                                                Band(
                                                    Ble(Cst(0), Var("b")),
                                                    Band(
                                                        Ble(Var("b"), Cst(0)),
                                                        Bnot(Beq(Var("b"), Cst(0))),
                                                    ),
                                                ),
                                            )
                                        ),
                                    ),
                                ),
                            ),
                        ),
                    ),
                ),
            ),
        )

        self.assertFalse(is_unsat(simplified_bug_path))

        # {}{old_a -> a}{old_b -> b}{a -> (a + b)}
        bug_store = NonEmptySymStore(
            "a",
            Add(Var("a"), Var("b")),
            NonEmptySymStore(
                "old_b", Var("b"), NonEmptySymStore("old_a", Var("a"), Id())
            ),
        )

        self.assertTrue(
            any(
                simplify_expr(elem.s[0]) == simplified_bug_path
                and elem.s[1] == bug_store
                for elem in bugs
            )
        )

    def test_true_assertion(self):
        prog = parse_imp("x = -1 ; assert x <= 0")
        result = peek(20, find_bugs(prog))
        self.assertTrue(
            all(
                not isinstance(elem, BugFound)
                or elem.s[0] == Band(Bcst(True), Bnot(Ble(Cst(-1), Cst(0))))  # unsat
                for elem in result
            )
        )

    def test_nested_loop(self):
        with open("../examples/nested_loop.imp", "r") as f:
            prog = parse_imp(f.read())

        result_breadth_first = peek(14, find_bugs(prog))
        result_depth_first = peek(100, find_bugs_depth_first(prog))

        def contains_satisfiable_bug(results):
            return any(
                isinstance(elem, BugFound) and not is_unsat(elem.s[0])
                for elem in results
            )

        self.assertTrue(contains_satisfiable_bug(result_breadth_first))
        self.assertFalse(contains_satisfiable_bug(result_depth_first))

    def test_integer_square_root(self):
        with open("../examples/integer_sqrt_buggy.imp", "r") as f:
            prog = parse_imp(f.read())

        result = peek(250, find_bugs(prog))

        def contains_satisfiable_bug(results):
            return any(
                isinstance(elem, BugFound) and not is_unsat(elem.s[0])
                # not `path ==> x <= 0`
                and not is_unsat(Bnot(Bor(Bnot(elem.s[0]), Ble(Var("x"), Cst(0)))))
                for elem in results
            )

        self.assertTrue(contains_satisfiable_bug(result))

    def test_factorial_correct(self):
        with open("../examples/factorial.imp", "r") as f:
            prog = parse_imp(f.read())

        assumption = parse_bexpr("n >= 2")
        result = peek(300, find_bugs_assuming(prog, assumption))

        assert not any(isinstance(elem, BugFound) for elem in result)

    def test_factorial_correct_exhaustive_with_assumption(self):
        with open("../examples/factorial.imp", "r") as f:
            prog = parse_imp(f.read())

        assumption = parse_bexpr("n >= 2 and n < 4")
        result = peek(200, find_bugs_assuming(prog, assumption))
        self.assertTrue(isinstance(result[-1], Finished))

    def test_factorial_buggy(self):
        with open("../examples/factorial_buggy.imp", "r") as f:
            prog = parse_imp(f.read())

        assumption = parse_bexpr("n >= 2")
        result = peek(300, find_bugs_assuming(prog, assumption))
        bugs: List[BugFound] = cast(
            List[BugFound], list(filter(BugFound.__instancecheck__, result))
        )

        # 2 <= n and 1 <= n and not (3 <= n)
        simplified_bug_path = Band(
            Ble(Cst(2), Var("n")),
            Band(Ble(Cst(1), Var("n")), Bnot(Ble(Cst(3), Var("n")))),
        )

        # {}{factorial -> 1}{i -> 1}{mul_result -> 1}{_i -> 1}{factorial -> 1}{i -> 2}
        # {mul_result -> 1}{_i -> 1}{mul_result -> 3}{_i -> 2}{factorial -> 3}{i -> 3}
        # {_i -> 2}{_m -> 3}{_d -> 0}{_d -> 1}{_m -> 1}
        simplified_store = NonEmptySymStore(
            "_m",
            Cst(1),
            NonEmptySymStore(
                "_d",
                Cst(1),
                NonEmptySymStore(
                    "_d",
                    Cst(0),
                    NonEmptySymStore(
                        "_m",
                        Cst(3),
                        NonEmptySymStore(
                            "_i",
                            Cst(2),
                            NonEmptySymStore(
                                "i",
                                Cst(3),
                                NonEmptySymStore(
                                    "factorial",
                                    Cst(3),
                                    NonEmptySymStore(
                                        "_i",
                                        Cst(2),
                                        NonEmptySymStore(
                                            "mul_result",
                                            Cst(3),
                                            NonEmptySymStore(
                                                "_i",
                                                Cst(1),
                                                NonEmptySymStore(
                                                    "mul_result",
                                                    Cst(1),
                                                    NonEmptySymStore(
                                                        "i",
                                                        Cst(2),
                                                        NonEmptySymStore(
                                                            "factorial",
                                                            Cst(1),
                                                            NonEmptySymStore(
                                                                "_i",
                                                                Cst(1),
                                                                NonEmptySymStore(
                                                                    "mul_result",
                                                                    Cst(1),
                                                                    NonEmptySymStore(
                                                                        "i",
                                                                        Cst(1),
                                                                        NonEmptySymStore(
                                                                            "factorial",
                                                                            Cst(1),
                                                                            Id(),
                                                                        ),
                                                                    ),
                                                                ),
                                                            ),
                                                        ),
                                                    ),
                                                ),
                                            ),
                                        ),
                                    ),
                                ),
                            ),
                        ),
                    ),
                ),
            ),
        )

        self.assertTrue(
            any(
                simplify_expr(bug.s[0]) == simplified_bug_path
                and simplify_store(bug.s[1]) == simplified_store
                for bug in bugs
            )
        )

    def Xtest_is_prime(self):
        # TODO
        with open("../examples/is_prime.imp", "r") as f:
            prog = parse_imp(f.read())

        for elem in peek(300, find_bugs(prog)):
            if not isinstance(elem, BugFound):
                continue

            if is_unsat(elem.s[0]):
                continue

            print(simplify_expr(elem.s[0]))


if __name__ == "__main__":
    unittest.main()



"""
(** A task list is a list of programs to be executed in a given symbolic state. *)
Definition tasks := list (sym_state * IMP).
"""
from dataclasses import dataclass
from typing import List, Tuple, Optional

from toolz import curry

from wise.helpers import is_unsat, simplify_expr, simplify_store
from wise.imp import (
    Imp,
    Seq,
    Bexpr,
    Skip,
    Aff,
    Err,
    Ite,
    Band,
    Bnot,
    Loop,
    is_error,
    Bcst,
)
from wise.streams import Stream, stream_map, scons
from wise.symex import SymState, SymStore, sym_update, sym_aeval, sym_beval, Id

# (** A task list is a list of programs to be executed in a given symbolic state. *)
# Definition tasks := list sym_state.
Tasks = List[SymState]

"""
(** A [status] corresponds to the current state of the bugfinding loop *)
Inductive status :=
  | BugFound (s : sym_state)
  | Pending
  | Finished.
"""


@dataclass(frozen=True)
class Status:
    pass


@dataclass(frozen=True)
class BugFound(Status):
    s: SymState


@dataclass(frozen=True)
class Pending(Status):
    pass


@dataclass(frozen=True)
class Finished(Status):
    pass


@curry
def then_do(q: Imp, task: SymState) -> SymState:
    """
    Definition then_do q '(((path, env), p) : sym_state) :=
      ((path, env), Seq p q).
    """
    path, env, p = task
    return path, env, Seq(p, q)


def exec_task(path: Bexpr, env: SymStore, prog: Imp) -> List[SymState]:
    """
    (** Executing a task according to [sym_step] *)
    Fixpoint exec_task path env prog : list sym_state :=
      match prog with
      | Skip => []
      | Aff x e =>
        [(path, sym_update env x (sym_aeval env e), Skip)]
      | Err => []
      | Ite b p1 p2 =>
        [
          (Band path (sym_beval env b), env, p1);
          (Band path (Bnot (sym_beval env b)), env, p2)
        ]
      | Loop b p =>
        [
          (Band path (sym_beval env b), env, Seq p (Loop b p));
          (Band path (Bnot (sym_beval env b)), env, Skip)
        ]
      | Seq Skip p2 => [(path, env, p2)]
      | Seq p1 p2 =>
        List.map (then_do p2) (exec_task path env p1)
      end.
    """

    match prog:
        case Skip():
            return []
        case Aff(x, e):
            return [(path, sym_update(env, x, sym_aeval(env, e)), Skip())]
        case Err():
            return []
        case Ite(b, p1, p2):
            return [
                (Band(path, sym_beval(env, b)), env, p1),
                (Band(path, Bnot(sym_beval(env, b))), env, p2),
            ]
        case Loop(b, p):
            return [
                (Band(path, sym_beval(env, b)), env, Seq(p, Loop(b, p))),
                (Band(path, Bnot(sym_beval(env, b))), env, Skip()),
            ]
        case Seq(Skip(), p2):
            return [(path, env, p2)]
        case Seq(p1, p2):
            return list(map(then_do(p2), exec_task(path, env, p1)))

    assert False


def run(l: Tasks) -> Stream[Optional[SymState]]:
    """
    (** Bugfinding loop: at every iteration, a task is choosen and then executed.
        If the execution results in an error, a [BugFound] message is
        emitted. If executing the task terminates in a state [s] without error,
        a message [Clear s] is emitted. If executing the task generates a list [l] of subtasks,
        then the loop signals that further computations are pending and add [l] to the worklist.
        Finally, if the task list is empty, the loop emits the token [Finished] continuously.
    *)
    CoFixpoint run (l : tasks) : stream (option sym_state) :=
      match l with
      | [] => scons None (run [])
      | ((path, env), prog)::next =>
        let next_next := exec_task path env prog in
        scons (Some (path, env, prog)) (run (next ++ next_next))
      end.
    """

    match l:
        case []:
            return scons(None, lambda: run([]))
        case [(path, env, prog), *next_]:
            if not is_unsat(path):
                next_next = exec_task(path, env, prog)
                return scons((path, env, prog), lambda: run(next_ + next_next))
            else:
                return run(next_)


def run_depth_first(l: Tasks) -> Stream[Optional[SymState]]:
    """
    A depth-first version of `run` as implemented by early symbolic executors like DART.
    """

    match l:
        case []:
            return scons(None, lambda: run_depth_first([]))
        case [(path, env, prog), *next_]:
            # Note: The only thing changed here is the order of `next_` and `next_next`
            #       in the recursive call to `run_depth_first`. This turns the procedure
            #       into a depth-first search.
            if not is_unsat(path):
                next_next = exec_task(path, env, prog)
                return scons(
                    (path, env, prog), lambda: run_depth_first(next_next + next_)
                )
            else:
                return run_depth_first(next_)


def display(n: Optional[SymState]) -> Status:
    """
    Definition display (st : option sym_state) : status :=
      match st with
      | None => Finished
      | Some (path, env, prog) =>
        if is_error prog then BugFound (path, env, prog)
        else Pending
      end.
    """
    match n:
        case None:
            return Finished()
        case (path, env, prog):
            return BugFound((path, env, prog)) if is_error(prog) else Pending()


def init(p: Imp) -> List[SymState]:
    """
    Definition init (p : IMP) : list sym_state :=
      [((Bcst true, id), p)].
    """
    return [(Bcst(True), Id(), p)]


def find_bugs(p: Imp) -> Stream[Status]:
    """
    Definition find_bugs (p : IMP) : stream status :=
      map display (run (init p)).
    """

    return stream_map(display, run(init(p)))


def find_bugs_depth_first(p: Imp, cond: Optional[Bexpr] = None) -> Stream[Status]:
    """
    A depth-first version of `find_bugs` using `run_depth_first`.
    """

    return stream_map(
        display, run_depth_first(init(p) if cond is None else init_assuming(p, cond))
    )


def init_assuming(p: Imp, cond: Bexpr) -> List[SymState]:
    """
    Definition init_assuming (p : IMP) (cond : bexpr) :=
      [(cond, id, p)].
    """

    return [(cond, Id(), p)]


def find_bugs_assuming(p: Imp, cond: Bexpr) -> Stream[Status]:
    """
    Definition find_bugs_assuming (p : IMP) (cond : bexpr) : stream status :=
      map display (run (init_assuming p cond)).
    """

    return stream_map(display, run(init_assuming(p, cond)))



from abc import abstractmethod
from dataclasses import dataclass

##############
### Syntax ###
##############
import z3

"""
Definition ident := string.
"""
Ident = str

Z = int

"""
Inductive aexpr : Type :=
  | Var : ident -> aexpr
  | Cst : Z -> aexpr
  | Add : aexpr -> aexpr -> aexpr
  | Sub : aexpr -> aexpr -> aexpr.
"""


@dataclass(frozen=True)
class Aexpr:
    @abstractmethod
    def to_smt(self) -> z3.ArithRef:
        raise NotImplementedError()


@dataclass(frozen=True)
class Var(Aexpr):
    ident: Ident

    def __str__(self):
        return self.ident

    def to_smt(self) -> z3.ArithRef:
        return z3.Int(self.ident)


@dataclass(frozen=True)
class Cst(Aexpr):
    num: Z

    def __str__(self):
        return str(self.num)

    def to_smt(self) -> z3.ArithRef:
        return z3.IntVal(self.num)


@dataclass(frozen=True)
class Add(Aexpr):
    left: Aexpr
    right: Aexpr

    def __str__(self):
        return f"({self.left} + {self.right})"

    def to_smt(self) -> z3.ArithRef:
        return self.left.to_smt() + self.right.to_smt()


@dataclass(frozen=True)
class Sub(Aexpr):
    left: Aexpr
    right: Aexpr

    def __str__(self):
        return f"({self.left} - {self.right})"

    def to_smt(self) -> z3.ArithRef:
        return self.left.to_smt() - self.right.to_smt()


"""
Inductive bexpr : Type :=
  | Bcst  : bool -> bexpr
  | Ble   : aexpr -> aexpr -> bexpr
  | Beq   : aexpr -> aexpr -> bexpr
  | Bnot  : bexpr -> bexpr
  | Band  : bexpr -> bexpr -> bexpr.
"""


@dataclass(frozen=True)
class Bexpr:
    @abstractmethod
    def to_smt(self) -> z3.BoolRef:
        raise NotImplementedError()


@dataclass(frozen=True)
class Bcst(Bexpr):
    b: bool

    def __str__(self):
        return str(self.b).lower()

    def to_smt(self) -> z3.BoolRef:
        return z3.BoolVal(self.b)


@dataclass(frozen=True)
class Ble(Bexpr):
    left: Aexpr
    right: Aexpr

    def __str__(self):
        return f"{self.left} <= {self.right}"

    def to_smt(self) -> z3.BoolRef:
        return self.left.to_smt() <= self.right.to_smt()


@dataclass(frozen=True)
class Beq(Bexpr):
    left: Aexpr
    right: Aexpr

    def __str__(self):
        return f"{self.left} == {self.right}"

    def to_smt(self) -> z3.BoolRef:
        return self.left.to_smt() == self.right.to_smt()


@dataclass(frozen=True)
class Bnot(Bexpr):
    expr: Bexpr

    def __str__(self):
        return f"not ({self.expr})"

    def to_smt(self) -> z3.BoolRef:
        return z3.Not(self.expr.to_smt())


@dataclass(frozen=True)
class Band(Bexpr):
    left: Bexpr
    right: Bexpr

    def __str__(self):
        return f"{self.left} and {self.right}"

    def to_smt(self) -> z3.BoolRef:
        return z3.And(self.left.to_smt(), self.right.to_smt())


def Bor(b1: Bexpr, b2: Bexpr) -> Bexpr:
    """
    Definition Bor (b1 b2 : bexpr) := Bnot (Band (Bnot b1) (Bnot b2)).
    """

    return Bnot(Band(Bnot(b1), Bnot(b2)))


"""
Inductive IMP : Type :=
  | Skip  : IMP
  | Ite   : bexpr -> IMP -> IMP -> IMP
  | Seq   : IMP -> IMP -> IMP
  | Aff   : string -> aexpr -> IMP
  | Err   : IMP
  | Loop  : bexpr -> IMP -> IMP.
"""


@dataclass(frozen=True)
class Imp:
    pass


@dataclass(frozen=True)
class Skip(Imp):
    pass

    def __str__(self):
        return "skip"


@dataclass(frozen=True)
class Ite(Imp):
    guard: Bexpr
    then_branch: Imp
    else_branch: Imp

    def __str__(self):
        return f"if {self.guard} then {self.then_branch} else {self.else_branch} fi"


@dataclass(frozen=True)
class Seq(Imp):
    first: Imp
    second: Imp

    def __str__(self):
        return f"{self.first} ; {self.second}"


@dataclass(frozen=True)
class Aff(Imp):
    var: Ident
    expr: Aexpr

    def __str__(self):
        return f"{self.var} = {self.expr}"


@dataclass(frozen=True)
class Err(Imp):
    pass

    def __str__(self):
        return "err"


@dataclass(frozen=True)
class Loop(Imp):
    guard: Bexpr
    body: Imp

    def __str__(self):
        return f"while {self.guard} do {self.body} od"


def Assert(c: Bexpr) -> Imp:
    """
    Definition Assert c := Ite c Skip Err.
    """
    return Ite(c, Skip(), Err())


def is_error(p: Imp) -> bool:
    """
    (** Executable check for erroneous state *)
    Fixpoint is_error (p : IMP) : bool :=
      match p with
      | Err => true
      | Seq p _ => is_error p
      | _ => false
      end.
    """
    match p:
        case Err():
            return True
        case Seq(p, _):
            return is_error(p)
        case _:
            return False



from typing import Dict, Optional, Tuple, cast, Generator, TypeVar, Union

import antlr4
from antlr4 import InputStream, ParserRuleContext, RuleContext
from antlr4.Token import CommonToken

from wise.helpers import eassert
from wise.imp import (
    Imp,
    Aexpr,
    Skip,
    Var,
    Aff,
    Cst,
    Add,
    Sub,
    Ite,
    Ble,
    Err,
    Bcst,
    Beq,
    Bnot,
    Band,
    Bor,
    Seq,
    Loop,
    Assert,
    Bexpr,
)
from wise.imp_language.ImpLanguageLexer import ImpLanguageLexer
from wise.imp_language.ImpLanguageListener import ImpLanguageListener
from wise.imp_language.ImpLanguageParser import ImpLanguageParser


class BailPrintErrorStrategy(antlr4.BailErrorStrategy):
    def recover(self, recognizer: antlr4.Parser, e: antlr4.RecognitionException):
        recognizer._errHandler.reportError(recognizer, e)
        super().recover(recognizer, e)


def parse_tree_text(elem: RuleContext | CommonToken) -> str:
    if isinstance(elem, CommonToken):
        return elem.text
    else:
        return elem.getText()


class ImpEmitter(ImpLanguageListener):
    def __init__(self):
        self.__result: Optional[Imp] = None

        self.aexprs: Dict[ParserRuleContext, Aexpr] = {}
        self.bexprs: Dict[ParserRuleContext, Bexpr] = {}
        self.progs: Dict[ParserRuleContext, Imp] = {}
        self.macros: Dict[str, Tuple[Tuple[Var], Imp]] = {}

    def result(self):
        assert self.__result is not None
        return self.__result

    def exitStart(self, ctx: ImpLanguageParser.StartContext):
        self.__result = self.progs[ctx.imp()]

    def exitVar(self, ctx: ImpLanguageParser.VarContext):
        self.aexprs[ctx] = Var(parse_tree_text(ctx.ID()))

    def exitCst(self, ctx: ImpLanguageParser.CstContext):
        self.aexprs[ctx] = Cst(int(parse_tree_text(ctx.NUM())))

    def exitAdd(self, ctx: ImpLanguageParser.AddContext):
        self.aexprs[ctx] = Add(
            self.aexprs[ctx.aexpr(0)], self.aexprs[ctx.aexpr(1)]
        )

    def exitSub(self, ctx: ImpLanguageParser.SubContext):
        self.aexprs[ctx] = Sub(
            self.aexprs[ctx.aexpr(0)], self.aexprs[ctx.aexpr(1)]
        )

    def exitApar(self, ctx: ImpLanguageParser.AparContext):
        self.aexprs[ctx] = self.aexprs[ctx.aexpr()]

    def exitBcst(self, ctx: ImpLanguageParser.BcstContext):
        self.bexprs[ctx] = Bcst(ctx.FALSE() is None)

    def exitBle(self, ctx: ImpLanguageParser.BleContext):
        self.bexprs[ctx] = Ble(
            self.aexprs[ctx.aexpr(0)], self.aexprs[ctx.aexpr(1)]
        )

    def exitBlt(self, ctx: ImpLanguageParser.BltContext):
        left = self.aexprs[ctx.aexpr(0)]
        right = self.aexprs[ctx.aexpr(1)]

        # a < b <==> a <= b and not a == b
        self.bexprs[ctx] = Band(Ble(left, right), Bnot(Beq(left, right)))

    def exitBge(self, ctx: ImpLanguageParser.BgeContext):
        left = self.aexprs[ctx.aexpr(0)]
        right = self.aexprs[ctx.aexpr(1)]

        # a >= b <==> b <= a
        self.bexprs[ctx] = Ble(right, left)

    def exitBgt(self, ctx: ImpLanguageParser.BgeContext):
        left = self.aexprs[ctx.aexpr(0)]
        right = self.aexprs[ctx.aexpr(1)]

        # a > b <==> b <= a and not a == b
        self.bexprs[ctx] = Band(Ble(right, left), Bnot(Beq(left, right)))

    def exitBeq(self, ctx: ImpLanguageParser.BeqContext):
        self.bexprs[ctx] = Beq(
            self.aexprs[ctx.aexpr(0)], self.aexprs[ctx.aexpr(1)]
        )

    def exitBnot(self, ctx: ImpLanguageParser.BnotContext):
        self.bexprs[ctx] = Bnot(self.bexprs[ctx.bexpr()])

    def exitBand(self, ctx: ImpLanguageParser.BandContext):
        self.bexprs[ctx] = Band(
            self.bexprs[ctx.bexpr(0)], self.bexprs[ctx.bexpr(1)]
        )

    def exitBor(self, ctx: ImpLanguageParser.BorContext):
        self.bexprs[ctx] = Bor(
            self.bexprs[ctx.bexpr(0)], self.bexprs[ctx.bexpr(1)]
        )

    def exitBpar(self, ctx: ImpLanguageParser.BparContext):
        self.bexprs[ctx] = self.bexprs[ctx.bexpr()]

    def exitSkip(self, ctx: ImpLanguageParser.SkipContext):
        self.progs[ctx] = Skip()

    def exitIte(self, ctx: ImpLanguageParser.IteContext):
        self.progs[ctx] = Ite(
            self.bexprs[ctx.bexpr()],
            self.progs[ctx.imp(0)],
            self.progs[ctx.imp(1)],
        )

    def exitSeq(self, ctx: ImpLanguageParser.SeqContext):
        self.progs[ctx] = Seq(self.progs[ctx.imp(0)], self.progs[ctx.imp(1)])

    def exitAff(self, ctx: ImpLanguageParser.AffContext):
        self.progs[ctx] = Aff(parse_tree_text(ctx.ID()), self.aexprs[ctx.aexpr()])

    def exitAssert(self, ctx: ImpLanguageParser.AssertContext):
        self.progs[ctx] = Assert(self.bexprs[ctx.bexpr()])

    def exitErr(self, ctx: ImpLanguageParser.ErrContext):
        self.progs[ctx] = Err()

    def exitLoop(self, ctx: ImpLanguageParser.LoopContext):
        self.progs[ctx] = Loop(self.bexprs[ctx.bexpr()], self.progs[ctx.imp()])

    def exitMacrodef(self, ctx: ImpLanguageParser.MacrodefContext):
        name = parse_tree_text(ctx.name)
        params = tuple([cast(Var, self.aexprs[var]) for var in ctx.aexpr()])
        assert all(isinstance(elem, Var) for elem in params)
        prog = self.progs[ctx.imp()]
        assert all(
            not isinstance(elem, Var) or elem in params for elem in walk_imp(prog)
        )

        self.macros[name] = (params, prog)

    def exitMacrocall(self, ctx: ImpLanguageParser.MacrocallContext):
        name = parse_tree_text(ctx.name)
        args = [self.aexprs[param] for param in ctx.aexpr()]
        assert name in self.macros
        assert len(args) == len(self.macros[name][0])
        self.progs[ctx] = replace_imp(
            self.macros[name][1], dict(zip(self.macros[name][0], args))
        )


T = TypeVar("T", bound=Union[Aexpr, Bexpr, Imp])


def replace_imp(orig: T, repl_dict: Dict) -> T:
    if orig in repl_dict:
        return repl_dict[orig]

    match orig:
        case Var(_) | Cst(_) | Bcst(_) | Skip() | Err():
            return orig
        case Bnot(a):
            return Bnot(replace_imp(a, repl_dict))
        case Add(a, b):
            return Add(replace_imp(a, repl_dict), replace_imp(b, repl_dict))
        case Sub(a, b):
            return Sub(replace_imp(a, repl_dict), replace_imp(b, repl_dict))
        case Ble(a, b):
            return Ble(replace_imp(a, repl_dict), replace_imp(b, repl_dict))
        case Beq(a, b):
            return Beq(replace_imp(a, repl_dict), replace_imp(b, repl_dict))
        case Band(a, b):
            return Band(replace_imp(a, repl_dict), replace_imp(b, repl_dict))
        case Seq(a, b):
            return Seq(replace_imp(a, repl_dict), replace_imp(b, repl_dict))
        case Aff(var, expr):
            new_lhs = replace_imp(Var(var), repl_dict)
            assert isinstance(new_lhs, Var)
            return Aff(new_lhs.ident, replace_imp(expr, repl_dict))
        case Ite(g, t, e):
            return Ite(
                replace_imp(g, repl_dict),
                replace_imp(t, repl_dict),
                replace_imp(e, repl_dict),
            )
        case Loop(g, b):
            return Loop(replace_imp(g, repl_dict), replace_imp(b, repl_dict))
        case _:
            assert False


def walk_imp(elem: Aexpr | Bexpr | Imp) -> Generator[Aexpr | Bexpr | Imp, None, None]:
    yield elem
    match elem:
        case Var(_) | Cst(_) | Bcst(_) | Skip() | Err():
            pass
        case Bnot(a):
            yield from walk_imp(a)
        case Add(a, b) | Sub(a, b) | Ble(a, b) | Beq(a, b) | Band(a, b) | Seq(a, b):
            yield from walk_imp(a)
            yield from walk_imp(b)
        case Aff(var, expr):
            yield Var(var)
            yield from walk_imp(expr)
        case Ite(g, t, e):
            yield from walk_imp(g)
            yield from walk_imp(t)
            yield from walk_imp(e)
        case Loop(g, b):
            yield from walk_imp(g)
            yield from walk_imp(b)
        case _:
            assert False


def parse_imp(inp: str) -> Imp:
    lexer = ImpLanguageLexer(InputStream(inp))
    parser = ImpLanguageParser(antlr4.CommonTokenStream(lexer))
    parser._errHandler = BailPrintErrorStrategy()
    imp_emitter = ImpEmitter()
    antlr4.ParseTreeWalker().walk(imp_emitter, parser.start())
    return imp_emitter.result()


def parse_bexpr(inp: str) -> Bexpr:
    lexer = ImpLanguageLexer(InputStream(inp))
    parser = ImpLanguageParser(antlr4.CommonTokenStream(lexer))
    parser._errHandler = BailPrintErrorStrategy()
    imp_emitter = ImpEmitter()
    ctx = parser.bexpr()
    antlr4.ParseTreeWalker().walk(imp_emitter, ctx)
    return imp_emitter.bexprs[ctx]



from typing import Generator, TypeVar, Callable, List

T = TypeVar("T")
S = TypeVar("S")

"""
CoInductive stream (A : Type) : Type :=
  | scons (x : A) (xs : stream A) : stream A.
"""
Stream = Generator[T, None, None]


def scons(x: T, xs: Callable[[], Stream[T]]) -> Stream[T]:
    yield x
    yield from xs()


def peek(n: int, s: Stream[T]) -> List[T]:
    """
    (** Peek the first [n] elements of a stream *)
    Fixpoint peek {A} (n : nat) (s : stream A) : list A :=
      match n, s with
      | 0, _ => []
      | S n, scons x xs => x::peek n xs
      end.
    """
    match n:
        case 0:
            return []
        case _:
            return [next(s)] + peek(n - 1, s)


def stream_map(f: Callable[[S], T], s: Stream[S]) -> Stream[T]:
    """
    (** Apply a function to a stream elementwise *)
    CoFixpoint map {A B} (f : A -> B) (s : stream A) : stream B :=
      match s with
      | scons x xs => scons (f x) (map f xs)
      end.
    """
    for elem in s:
        yield f(elem)



# Generated from ImpLanguage.g4 by ANTLR 4.11.1
# encoding: utf-8
from antlr4 import *
from io import StringIO
import sys
if sys.version_info[1] > 5:
	from typing import TextIO
else:
	from typing.io import TextIO

def serializedATN():
    return [
        4,1,34,145,2,0,7,0,2,1,7,1,2,2,7,2,2,3,7,3,2,4,7,4,1,0,5,0,12,8,
        0,10,0,12,0,15,9,0,1,0,1,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,3,1,26,8,
        1,1,1,1,1,1,1,1,1,1,1,1,1,5,1,34,8,1,10,1,12,1,37,9,1,1,2,1,2,1,
        2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,
        2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,3,2,67,8,2,1,2,1,2,1,2,1,2,
        1,2,1,2,5,2,75,8,2,10,2,12,2,78,9,2,1,3,1,3,1,3,1,3,1,3,1,3,1,3,
        1,3,1,3,1,3,1,3,1,3,1,3,1,3,1,3,1,3,1,3,1,3,1,3,1,3,1,3,1,3,1,3,
        1,3,1,3,1,3,1,3,5,3,107,8,3,10,3,12,3,110,9,3,1,3,1,3,3,3,114,8,
        3,3,3,116,8,3,1,3,1,3,1,3,5,3,121,8,3,10,3,12,3,124,9,3,1,4,1,4,
        1,4,1,4,1,4,1,4,5,4,132,8,4,10,4,12,4,135,9,4,1,4,1,4,3,4,139,8,
        4,1,4,1,4,1,4,1,4,1,4,0,3,2,4,6,5,0,2,4,6,8,0,1,1,0,29,30,164,0,
        13,1,0,0,0,2,25,1,0,0,0,4,66,1,0,0,0,6,115,1,0,0,0,8,125,1,0,0,0,
        10,12,3,8,4,0,11,10,1,0,0,0,12,15,1,0,0,0,13,11,1,0,0,0,13,14,1,
        0,0,0,14,16,1,0,0,0,15,13,1,0,0,0,16,17,3,6,3,0,17,1,1,0,0,0,18,
        19,6,1,-1,0,19,26,5,32,0,0,20,26,5,31,0,0,21,22,5,3,0,0,22,23,3,
        2,1,0,23,24,5,4,0,0,24,26,1,0,0,0,25,18,1,0,0,0,25,20,1,0,0,0,25,
        21,1,0,0,0,26,35,1,0,0,0,27,28,10,3,0,0,28,29,5,1,0,0,29,34,3,2,
        1,4,30,31,10,2,0,0,31,32,5,2,0,0,32,34,3,2,1,3,33,27,1,0,0,0,33,
        30,1,0,0,0,34,37,1,0,0,0,35,33,1,0,0,0,35,36,1,0,0,0,36,3,1,0,0,
        0,37,35,1,0,0,0,38,39,6,2,-1,0,39,67,7,0,0,0,40,41,3,2,1,0,41,42,
        5,5,0,0,42,43,3,2,1,0,43,67,1,0,0,0,44,45,3,2,1,0,45,46,5,6,0,0,
        46,47,3,2,1,0,47,67,1,0,0,0,48,49,3,2,1,0,49,50,5,7,0,0,50,51,3,
        2,1,0,51,67,1,0,0,0,52,53,3,2,1,0,53,54,5,8,0,0,54,55,3,2,1,0,55,
        67,1,0,0,0,56,57,3,2,1,0,57,58,5,9,0,0,58,59,3,2,1,0,59,67,1,0,0,
        0,60,61,5,10,0,0,61,67,3,4,2,4,62,63,5,3,0,0,63,64,3,4,2,0,64,65,
        5,4,0,0,65,67,1,0,0,0,66,38,1,0,0,0,66,40,1,0,0,0,66,44,1,0,0,0,
        66,48,1,0,0,0,66,52,1,0,0,0,66,56,1,0,0,0,66,60,1,0,0,0,66,62,1,
        0,0,0,67,76,1,0,0,0,68,69,10,3,0,0,69,70,5,11,0,0,70,75,3,4,2,4,
        71,72,10,2,0,0,72,73,5,12,0,0,73,75,3,4,2,3,74,68,1,0,0,0,74,71,
        1,0,0,0,75,78,1,0,0,0,76,74,1,0,0,0,76,77,1,0,0,0,77,5,1,0,0,0,78,
        76,1,0,0,0,79,80,6,3,-1,0,80,116,5,13,0,0,81,116,5,14,0,0,82,83,
        5,32,0,0,83,84,5,15,0,0,84,116,3,2,1,0,85,86,5,16,0,0,86,116,3,4,
        2,0,87,88,5,17,0,0,88,89,3,4,2,0,89,90,5,18,0,0,90,91,3,6,3,0,91,
        92,5,19,0,0,92,93,3,6,3,0,93,94,5,20,0,0,94,116,1,0,0,0,95,96,5,
        21,0,0,96,97,3,4,2,0,97,98,5,22,0,0,98,99,3,6,3,0,99,100,5,23,0,
        0,100,116,1,0,0,0,101,113,5,32,0,0,102,103,5,3,0,0,103,108,3,2,1,
        0,104,105,5,24,0,0,105,107,3,2,1,0,106,104,1,0,0,0,107,110,1,0,0,
        0,108,106,1,0,0,0,108,109,1,0,0,0,109,111,1,0,0,0,110,108,1,0,0,
        0,111,112,5,4,0,0,112,114,1,0,0,0,113,102,1,0,0,0,113,114,1,0,0,
        0,114,116,1,0,0,0,115,79,1,0,0,0,115,81,1,0,0,0,115,82,1,0,0,0,115,
        85,1,0,0,0,115,87,1,0,0,0,115,95,1,0,0,0,115,101,1,0,0,0,116,122,
        1,0,0,0,117,118,10,1,0,0,118,119,5,25,0,0,119,121,3,6,3,2,120,117,
        1,0,0,0,121,124,1,0,0,0,122,120,1,0,0,0,122,123,1,0,0,0,123,7,1,
        0,0,0,124,122,1,0,0,0,125,126,5,26,0,0,126,138,5,32,0,0,127,128,
        5,3,0,0,128,133,3,2,1,0,129,130,5,24,0,0,130,132,3,2,1,0,131,129,
        1,0,0,0,132,135,1,0,0,0,133,131,1,0,0,0,133,134,1,0,0,0,134,136,
        1,0,0,0,135,133,1,0,0,0,136,137,5,4,0,0,137,139,1,0,0,0,138,127,
        1,0,0,0,138,139,1,0,0,0,139,140,1,0,0,0,140,141,5,27,0,0,141,142,
        3,6,3,0,142,143,5,28,0,0,143,9,1,0,0,0,13,13,25,33,35,66,74,76,108,
        113,115,122,133,138
    ]

class ImpLanguageParser ( Parser ):

    grammarFileName = "ImpLanguage.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [ DFA(ds, i) for i, ds in enumerate(atn.decisionToState) ]

    sharedContextCache = PredictionContextCache()

    literalNames = [ "<INVALID>", "'+'", "'-'", "'('", "')'", "'<='", "'<'", 
                     "'>='", "'>'", "'=='", "'not'", "'and'", "'or'", "'skip'", 
                     "'fail'", "'='", "'assert'", "'if'", "'then'", "'else'", 
                     "'fi'", "'while'", "'do'", "'od'", "','", "';'", "'macro'", 
                     "'begin'", "'end'", "'true'", "'false'" ]

    symbolicNames = [ "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "TRUE", "FALSE", "NUM", "ID", "WS", "COMMENT" ]

    RULE_start = 0
    RULE_aexpr = 1
    RULE_bexpr = 2
    RULE_imp = 3
    RULE_macrodef = 4

    ruleNames =  [ "start", "aexpr", "bexpr", "imp", "macrodef" ]

    EOF = Token.EOF
    T__0=1
    T__1=2
    T__2=3
    T__3=4
    T__4=5
    T__5=6
    T__6=7
    T__7=8
    T__8=9
    T__9=10
    T__10=11
    T__11=12
    T__12=13
    T__13=14
    T__14=15
    T__15=16
    T__16=17
    T__17=18
    T__18=19
    T__19=20
    T__20=21
    T__21=22
    T__22=23
    T__23=24
    T__24=25
    T__25=26
    T__26=27
    T__27=28
    TRUE=29
    FALSE=30
    NUM=31
    ID=32
    WS=33
    COMMENT=34

    def __init__(self, input:TokenStream, output:TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.11.1")
        self._interp = ParserATNSimulator(self, self.atn, self.decisionsToDFA, self.sharedContextCache)
        self._predicates = None




    class StartContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def imp(self):
            return self.getTypedRuleContext(ImpLanguageParser.ImpContext,0)


        def macrodef(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(ImpLanguageParser.MacrodefContext)
            else:
                return self.getTypedRuleContext(ImpLanguageParser.MacrodefContext,i)


        def getRuleIndex(self):
            return ImpLanguageParser.RULE_start

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterStart" ):
                listener.enterStart(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitStart" ):
                listener.exitStart(self)




    def start(self):

        localctx = ImpLanguageParser.StartContext(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_start)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 13
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==26:
                self.state = 10
                self.macrodef()
                self.state = 15
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 16
            self.imp(0)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class AexprContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return ImpLanguageParser.RULE_aexpr

     
        def copyFrom(self, ctx:ParserRuleContext):
            super().copyFrom(ctx)


    class AddContext(AexprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a ImpLanguageParser.AexprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def aexpr(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(ImpLanguageParser.AexprContext)
            else:
                return self.getTypedRuleContext(ImpLanguageParser.AexprContext,i)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterAdd" ):
                listener.enterAdd(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitAdd" ):
                listener.exitAdd(self)


    class SubContext(AexprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a ImpLanguageParser.AexprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def aexpr(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(ImpLanguageParser.AexprContext)
            else:
                return self.getTypedRuleContext(ImpLanguageParser.AexprContext,i)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterSub" ):
                listener.enterSub(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitSub" ):
                listener.exitSub(self)


    class CstContext(AexprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a ImpLanguageParser.AexprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def NUM(self):
            return self.getToken(ImpLanguageParser.NUM, 0)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterCst" ):
                listener.enterCst(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitCst" ):
                listener.exitCst(self)


    class VarContext(AexprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a ImpLanguageParser.AexprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def ID(self):
            return self.getToken(ImpLanguageParser.ID, 0)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterVar" ):
                listener.enterVar(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitVar" ):
                listener.exitVar(self)


    class AparContext(AexprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a ImpLanguageParser.AexprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def aexpr(self):
            return self.getTypedRuleContext(ImpLanguageParser.AexprContext,0)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterApar" ):
                listener.enterApar(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitApar" ):
                listener.exitApar(self)



    def aexpr(self, _p:int=0):
        _parentctx = self._ctx
        _parentState = self.state
        localctx = ImpLanguageParser.AexprContext(self, self._ctx, _parentState)
        _prevctx = localctx
        _startState = 2
        self.enterRecursionRule(localctx, 2, self.RULE_aexpr, _p)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 25
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [32]:
                localctx = ImpLanguageParser.VarContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx

                self.state = 19
                self.match(ImpLanguageParser.ID)
                pass
            elif token in [31]:
                localctx = ImpLanguageParser.CstContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 20
                self.match(ImpLanguageParser.NUM)
                pass
            elif token in [3]:
                localctx = ImpLanguageParser.AparContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 21
                self.match(ImpLanguageParser.T__2)
                self.state = 22
                self.aexpr(0)
                self.state = 23
                self.match(ImpLanguageParser.T__3)
                pass
            else:
                raise NoViableAltException(self)

            self._ctx.stop = self._input.LT(-1)
            self.state = 35
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input,3,self._ctx)
            while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                if _alt==1:
                    if self._parseListeners is not None:
                        self.triggerExitRuleEvent()
                    _prevctx = localctx
                    self.state = 33
                    self._errHandler.sync(self)
                    la_ = self._interp.adaptivePredict(self._input,2,self._ctx)
                    if la_ == 1:
                        localctx = ImpLanguageParser.AddContext(self, ImpLanguageParser.AexprContext(self, _parentctx, _parentState))
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_aexpr)
                        self.state = 27
                        if not self.precpred(self._ctx, 3):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(self, "self.precpred(self._ctx, 3)")
                        self.state = 28
                        self.match(ImpLanguageParser.T__0)
                        self.state = 29
                        self.aexpr(4)
                        pass

                    elif la_ == 2:
                        localctx = ImpLanguageParser.SubContext(self, ImpLanguageParser.AexprContext(self, _parentctx, _parentState))
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_aexpr)
                        self.state = 30
                        if not self.precpred(self._ctx, 2):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(self, "self.precpred(self._ctx, 2)")
                        self.state = 31
                        self.match(ImpLanguageParser.T__1)
                        self.state = 32
                        self.aexpr(3)
                        pass

             
                self.state = 37
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,3,self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.unrollRecursionContexts(_parentctx)
        return localctx


    class BexprContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return ImpLanguageParser.RULE_bexpr

     
        def copyFrom(self, ctx:ParserRuleContext):
            super().copyFrom(ctx)


    class BgeContext(BexprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a ImpLanguageParser.BexprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def aexpr(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(ImpLanguageParser.AexprContext)
            else:
                return self.getTypedRuleContext(ImpLanguageParser.AexprContext,i)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterBge" ):
                listener.enterBge(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitBge" ):
                listener.exitBge(self)


    class BparContext(BexprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a ImpLanguageParser.BexprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def bexpr(self):
            return self.getTypedRuleContext(ImpLanguageParser.BexprContext,0)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterBpar" ):
                listener.enterBpar(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitBpar" ):
                listener.exitBpar(self)


    class BorContext(BexprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a ImpLanguageParser.BexprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def bexpr(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(ImpLanguageParser.BexprContext)
            else:
                return self.getTypedRuleContext(ImpLanguageParser.BexprContext,i)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterBor" ):
                listener.enterBor(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitBor" ):
                listener.exitBor(self)


    class BleContext(BexprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a ImpLanguageParser.BexprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def aexpr(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(ImpLanguageParser.AexprContext)
            else:
                return self.getTypedRuleContext(ImpLanguageParser.AexprContext,i)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterBle" ):
                listener.enterBle(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitBle" ):
                listener.exitBle(self)


    class BltContext(BexprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a ImpLanguageParser.BexprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def aexpr(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(ImpLanguageParser.AexprContext)
            else:
                return self.getTypedRuleContext(ImpLanguageParser.AexprContext,i)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterBlt" ):
                listener.enterBlt(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitBlt" ):
                listener.exitBlt(self)


    class BandContext(BexprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a ImpLanguageParser.BexprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def bexpr(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(ImpLanguageParser.BexprContext)
            else:
                return self.getTypedRuleContext(ImpLanguageParser.BexprContext,i)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterBand" ):
                listener.enterBand(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitBand" ):
                listener.exitBand(self)


    class BcstContext(BexprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a ImpLanguageParser.BexprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def TRUE(self):
            return self.getToken(ImpLanguageParser.TRUE, 0)
        def FALSE(self):
            return self.getToken(ImpLanguageParser.FALSE, 0)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterBcst" ):
                listener.enterBcst(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitBcst" ):
                listener.exitBcst(self)


    class BgtContext(BexprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a ImpLanguageParser.BexprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def aexpr(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(ImpLanguageParser.AexprContext)
            else:
                return self.getTypedRuleContext(ImpLanguageParser.AexprContext,i)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterBgt" ):
                listener.enterBgt(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitBgt" ):
                listener.exitBgt(self)


    class BeqContext(BexprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a ImpLanguageParser.BexprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def aexpr(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(ImpLanguageParser.AexprContext)
            else:
                return self.getTypedRuleContext(ImpLanguageParser.AexprContext,i)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterBeq" ):
                listener.enterBeq(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitBeq" ):
                listener.exitBeq(self)


    class BnotContext(BexprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a ImpLanguageParser.BexprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def bexpr(self):
            return self.getTypedRuleContext(ImpLanguageParser.BexprContext,0)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterBnot" ):
                listener.enterBnot(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitBnot" ):
                listener.exitBnot(self)



    def bexpr(self, _p:int=0):
        _parentctx = self._ctx
        _parentState = self.state
        localctx = ImpLanguageParser.BexprContext(self, self._ctx, _parentState)
        _prevctx = localctx
        _startState = 4
        self.enterRecursionRule(localctx, 4, self.RULE_bexpr, _p)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 66
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,4,self._ctx)
            if la_ == 1:
                localctx = ImpLanguageParser.BcstContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx

                self.state = 39
                _la = self._input.LA(1)
                if not(_la==29 or _la==30):
                    self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()
                pass

            elif la_ == 2:
                localctx = ImpLanguageParser.BleContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 40
                self.aexpr(0)
                self.state = 41
                self.match(ImpLanguageParser.T__4)
                self.state = 42
                self.aexpr(0)
                pass

            elif la_ == 3:
                localctx = ImpLanguageParser.BltContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 44
                self.aexpr(0)
                self.state = 45
                self.match(ImpLanguageParser.T__5)
                self.state = 46
                self.aexpr(0)
                pass

            elif la_ == 4:
                localctx = ImpLanguageParser.BgeContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 48
                self.aexpr(0)
                self.state = 49
                self.match(ImpLanguageParser.T__6)
                self.state = 50
                self.aexpr(0)
                pass

            elif la_ == 5:
                localctx = ImpLanguageParser.BgtContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 52
                self.aexpr(0)
                self.state = 53
                self.match(ImpLanguageParser.T__7)
                self.state = 54
                self.aexpr(0)
                pass

            elif la_ == 6:
                localctx = ImpLanguageParser.BeqContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 56
                self.aexpr(0)
                self.state = 57
                self.match(ImpLanguageParser.T__8)
                self.state = 58
                self.aexpr(0)
                pass

            elif la_ == 7:
                localctx = ImpLanguageParser.BnotContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 60
                self.match(ImpLanguageParser.T__9)
                self.state = 61
                self.bexpr(4)
                pass

            elif la_ == 8:
                localctx = ImpLanguageParser.BparContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 62
                self.match(ImpLanguageParser.T__2)
                self.state = 63
                self.bexpr(0)
                self.state = 64
                self.match(ImpLanguageParser.T__3)
                pass


            self._ctx.stop = self._input.LT(-1)
            self.state = 76
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input,6,self._ctx)
            while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                if _alt==1:
                    if self._parseListeners is not None:
                        self.triggerExitRuleEvent()
                    _prevctx = localctx
                    self.state = 74
                    self._errHandler.sync(self)
                    la_ = self._interp.adaptivePredict(self._input,5,self._ctx)
                    if la_ == 1:
                        localctx = ImpLanguageParser.BandContext(self, ImpLanguageParser.BexprContext(self, _parentctx, _parentState))
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_bexpr)
                        self.state = 68
                        if not self.precpred(self._ctx, 3):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(self, "self.precpred(self._ctx, 3)")
                        self.state = 69
                        self.match(ImpLanguageParser.T__10)
                        self.state = 70
                        self.bexpr(4)
                        pass

                    elif la_ == 2:
                        localctx = ImpLanguageParser.BorContext(self, ImpLanguageParser.BexprContext(self, _parentctx, _parentState))
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_bexpr)
                        self.state = 71
                        if not self.precpred(self._ctx, 2):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(self, "self.precpred(self._ctx, 2)")
                        self.state = 72
                        self.match(ImpLanguageParser.T__11)
                        self.state = 73
                        self.bexpr(3)
                        pass

             
                self.state = 78
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,6,self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.unrollRecursionContexts(_parentctx)
        return localctx


    class ImpContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return ImpLanguageParser.RULE_imp

     
        def copyFrom(self, ctx:ParserRuleContext):
            super().copyFrom(ctx)


    class AffContext(ImpContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a ImpLanguageParser.ImpContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def ID(self):
            return self.getToken(ImpLanguageParser.ID, 0)
        def aexpr(self):
            return self.getTypedRuleContext(ImpLanguageParser.AexprContext,0)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterAff" ):
                listener.enterAff(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitAff" ):
                listener.exitAff(self)


    class ErrContext(ImpContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a ImpLanguageParser.ImpContext
            super().__init__(parser)
            self.copyFrom(ctx)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterErr" ):
                listener.enterErr(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitErr" ):
                listener.exitErr(self)


    class AssertContext(ImpContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a ImpLanguageParser.ImpContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def bexpr(self):
            return self.getTypedRuleContext(ImpLanguageParser.BexprContext,0)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterAssert" ):
                listener.enterAssert(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitAssert" ):
                listener.exitAssert(self)


    class LoopContext(ImpContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a ImpLanguageParser.ImpContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def bexpr(self):
            return self.getTypedRuleContext(ImpLanguageParser.BexprContext,0)

        def imp(self):
            return self.getTypedRuleContext(ImpLanguageParser.ImpContext,0)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterLoop" ):
                listener.enterLoop(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitLoop" ):
                listener.exitLoop(self)


    class SkipContext(ImpContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a ImpLanguageParser.ImpContext
            super().__init__(parser)
            self.copyFrom(ctx)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterSkip" ):
                listener.enterSkip(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitSkip" ):
                listener.exitSkip(self)


    class IteContext(ImpContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a ImpLanguageParser.ImpContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def bexpr(self):
            return self.getTypedRuleContext(ImpLanguageParser.BexprContext,0)

        def imp(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(ImpLanguageParser.ImpContext)
            else:
                return self.getTypedRuleContext(ImpLanguageParser.ImpContext,i)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterIte" ):
                listener.enterIte(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitIte" ):
                listener.exitIte(self)


    class MacrocallContext(ImpContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a ImpLanguageParser.ImpContext
            super().__init__(parser)
            self.name = None # Token
            self.copyFrom(ctx)

        def ID(self):
            return self.getToken(ImpLanguageParser.ID, 0)
        def aexpr(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(ImpLanguageParser.AexprContext)
            else:
                return self.getTypedRuleContext(ImpLanguageParser.AexprContext,i)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterMacrocall" ):
                listener.enterMacrocall(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitMacrocall" ):
                listener.exitMacrocall(self)


    class SeqContext(ImpContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a ImpLanguageParser.ImpContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def imp(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(ImpLanguageParser.ImpContext)
            else:
                return self.getTypedRuleContext(ImpLanguageParser.ImpContext,i)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterSeq" ):
                listener.enterSeq(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitSeq" ):
                listener.exitSeq(self)



    def imp(self, _p:int=0):
        _parentctx = self._ctx
        _parentState = self.state
        localctx = ImpLanguageParser.ImpContext(self, self._ctx, _parentState)
        _prevctx = localctx
        _startState = 6
        self.enterRecursionRule(localctx, 6, self.RULE_imp, _p)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 115
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,9,self._ctx)
            if la_ == 1:
                localctx = ImpLanguageParser.SkipContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx

                self.state = 80
                self.match(ImpLanguageParser.T__12)
                pass

            elif la_ == 2:
                localctx = ImpLanguageParser.ErrContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 81
                self.match(ImpLanguageParser.T__13)
                pass

            elif la_ == 3:
                localctx = ImpLanguageParser.AffContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 82
                self.match(ImpLanguageParser.ID)
                self.state = 83
                self.match(ImpLanguageParser.T__14)
                self.state = 84
                self.aexpr(0)
                pass

            elif la_ == 4:
                localctx = ImpLanguageParser.AssertContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 85
                self.match(ImpLanguageParser.T__15)
                self.state = 86
                self.bexpr(0)
                pass

            elif la_ == 5:
                localctx = ImpLanguageParser.IteContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 87
                self.match(ImpLanguageParser.T__16)
                self.state = 88
                self.bexpr(0)
                self.state = 89
                self.match(ImpLanguageParser.T__17)
                self.state = 90
                self.imp(0)
                self.state = 91
                self.match(ImpLanguageParser.T__18)
                self.state = 92
                self.imp(0)
                self.state = 93
                self.match(ImpLanguageParser.T__19)
                pass

            elif la_ == 6:
                localctx = ImpLanguageParser.LoopContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 95
                self.match(ImpLanguageParser.T__20)
                self.state = 96
                self.bexpr(0)
                self.state = 97
                self.match(ImpLanguageParser.T__21)
                self.state = 98
                self.imp(0)
                self.state = 99
                self.match(ImpLanguageParser.T__22)
                pass

            elif la_ == 7:
                localctx = ImpLanguageParser.MacrocallContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 101
                localctx.name = self.match(ImpLanguageParser.ID)
                self.state = 113
                self._errHandler.sync(self)
                la_ = self._interp.adaptivePredict(self._input,8,self._ctx)
                if la_ == 1:
                    self.state = 102
                    self.match(ImpLanguageParser.T__2)
                    self.state = 103
                    self.aexpr(0)
                    self.state = 108
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)
                    while _la==24:
                        self.state = 104
                        self.match(ImpLanguageParser.T__23)
                        self.state = 105
                        self.aexpr(0)
                        self.state = 110
                        self._errHandler.sync(self)
                        _la = self._input.LA(1)

                    self.state = 111
                    self.match(ImpLanguageParser.T__3)


                pass


            self._ctx.stop = self._input.LT(-1)
            self.state = 122
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input,10,self._ctx)
            while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                if _alt==1:
                    if self._parseListeners is not None:
                        self.triggerExitRuleEvent()
                    _prevctx = localctx
                    localctx = ImpLanguageParser.SeqContext(self, ImpLanguageParser.ImpContext(self, _parentctx, _parentState))
                    self.pushNewRecursionContext(localctx, _startState, self.RULE_imp)
                    self.state = 117
                    if not self.precpred(self._ctx, 1):
                        from antlr4.error.Errors import FailedPredicateException
                        raise FailedPredicateException(self, "self.precpred(self._ctx, 1)")
                    self.state = 118
                    self.match(ImpLanguageParser.T__24)
                    self.state = 119
                    self.imp(2) 
                self.state = 124
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,10,self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.unrollRecursionContexts(_parentctx)
        return localctx


    class MacrodefContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser
            self.name = None # Token

        def imp(self):
            return self.getTypedRuleContext(ImpLanguageParser.ImpContext,0)


        def ID(self):
            return self.getToken(ImpLanguageParser.ID, 0)

        def aexpr(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(ImpLanguageParser.AexprContext)
            else:
                return self.getTypedRuleContext(ImpLanguageParser.AexprContext,i)


        def getRuleIndex(self):
            return ImpLanguageParser.RULE_macrodef

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterMacrodef" ):
                listener.enterMacrodef(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitMacrodef" ):
                listener.exitMacrodef(self)




    def macrodef(self):

        localctx = ImpLanguageParser.MacrodefContext(self, self._ctx, self.state)
        self.enterRule(localctx, 8, self.RULE_macrodef)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 125
            self.match(ImpLanguageParser.T__25)
            self.state = 126
            localctx.name = self.match(ImpLanguageParser.ID)
            self.state = 138
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==3:
                self.state = 127
                self.match(ImpLanguageParser.T__2)
                self.state = 128
                self.aexpr(0)
                self.state = 133
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while _la==24:
                    self.state = 129
                    self.match(ImpLanguageParser.T__23)
                    self.state = 130
                    self.aexpr(0)
                    self.state = 135
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)

                self.state = 136
                self.match(ImpLanguageParser.T__3)


            self.state = 140
            self.match(ImpLanguageParser.T__26)
            self.state = 141
            self.imp(0)
            self.state = 142
            self.match(ImpLanguageParser.T__27)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx



    def sempred(self, localctx:RuleContext, ruleIndex:int, predIndex:int):
        if self._predicates == None:
            self._predicates = dict()
        self._predicates[1] = self.aexpr_sempred
        self._predicates[2] = self.bexpr_sempred
        self._predicates[3] = self.imp_sempred
        pred = self._predicates.get(ruleIndex, None)
        if pred is None:
            raise Exception("No predicate with index:" + str(ruleIndex))
        else:
            return pred(localctx, predIndex)

    def aexpr_sempred(self, localctx:AexprContext, predIndex:int):
            if predIndex == 0:
                return self.precpred(self._ctx, 3)
         

            if predIndex == 1:
                return self.precpred(self._ctx, 2)
         

    def bexpr_sempred(self, localctx:BexprContext, predIndex:int):
            if predIndex == 2:
                return self.precpred(self._ctx, 3)
         

            if predIndex == 3:
                return self.precpred(self._ctx, 2)
         

    def imp_sempred(self, localctx:ImpContext, predIndex:int):
            if predIndex == 4:
                return self.precpred(self._ctx, 1)
         







import unittest

from wise.imp import *
from wise.parser import parse_imp, parse_bexpr


class TestParser(unittest.TestCase):
    def test_simple_programs(self):
        self.assertEqual(Skip(), parse_imp("skip"))
        self.assertEqual(
            Aff("x", Add(Cst(17), Cst(23))),
            parse_imp("x = 17 + 23"),
        )
        self.assertEqual(
            Aff("x", Sub(Cst(17), Add(Cst(23), Var("y")))),
            parse_imp("x = 17 - (23 + y)"),
        )
        self.assertEqual(
            Aff("x", Add(Sub(Cst(17), Cst(23)), Var("y"))),
            parse_imp("x = (17 - 23) + y"),
        )
        self.assertEqual(
            Ite(
                Band(Bnot(Beq(Var("x"), Var("y"))), Beq(Var("x"), Var("y"))),
                Skip(),
                Err(),
            ),
            parse_imp("if not x == y and x == y then skip else fail fi"),
        )
        self.assertEqual(
            Ite(
                Bnot(Band(Beq(Var("x"), Var("y")), Beq(Var("x"), Cst(17)))),
                Skip(),
                Err(),
            ),
            parse_imp("if not(x == y and x == 17) then skip else fail fi"),
        )
        self.assertEqual(
            Seq(Seq(Aff("a", Var("b")), Aff("c", Cst(17))), Err()),
            parse_imp("a = b ; c = 17 ; fail"),
        )
        self.assertEqual(Loop(Bcst(True), Skip()), parse_imp("while true do skip od"))
        self.assertEqual(Ite(Bcst(False), Skip(), Err()), parse_imp("assert false"))

    def test_nonparametric_macro(self):
        result = parse_imp("macro useless begin skip end useless")
        self.assertEqual(Skip(), result)

    def test_macro_calls_in_macro(self):
        result = parse_imp(
            """
        macro useless 
        begin skip end
         
        macro useless2 
        begin useless end
        
        useless2"""
        )

        self.assertEqual(Skip(), result)

    def test_subadd_macro(self):
        result = parse_imp(
            """
        macro subadd1(x, y)
        begin
          x = x - y ;
          x = x + 1
        end
        subadd1(a, b);
        a = 0"""
        )

        expected = Seq(
            Seq(Aff("a", Sub(Var("a"), Var("b"))), Aff("a", Add(Var("a"), Cst(1)))),
            Aff("a", Cst(0)),
        )

        self.assertEqual(expected, result)

    def test_parse_bexpr(self):
        result = parse_bexpr("a >= b and not a==b")
        expected = Band(Ble(Var("b"), Var("a")), Bnot(Beq(Var("a"), Var("b"))))
        self.assertEqual(expected, result)


if __name__ == "__main__":
    unittest.main()



import argparse
import sys

import z3
from antlr4.error.Errors import ParseCancellationException

import wise
from wise.bugfinder import find_bugs, BugFound, find_bugs_depth_first, \
    find_bugs_assuming
from wise.helpers import is_unsat, simplify_expr
from wise.parser import parse_imp, parse_bexpr

# Exit Codes
USAGE_ERROR = 2
DATA_FORMAT_ERROR = 65


def main():
    parser = argparse.ArgumentParser(
        prog="wise",
        description="A Python implementation of the verified symbolic executor WiSE "
        + f"(v{wise.__version__})",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    parser.add_argument(
        "-n",
        "--nodes",
        type=int,
        default=50,
        help="the number of symbolic states to explore",
    )

    parser.add_argument(
        "--depth-first",
        type=bool,
        action=argparse.BooleanOptionalAction,
        default=False,
        help="run symbolic execution in depth-first mode (not as in original WiSE)",
    )

    parser.add_argument(
        "--simplify",
        type=bool,
        action=argparse.BooleanOptionalAction,
        default=True,
        help="simplify output path conditions",
    )

    parser.add_argument(
        "--prune-early",
        type=bool,
        action=argparse.BooleanOptionalAction,
        default=True,
        help="check intermediate states for satisfiability and prune unsatisfiable ones",
    )

    parser.add_argument(
        "-a",
        "--assumption",
        required=False,
        help="an optional initial assumption",
    )

    parser.add_argument(
        "file",
        metavar="FILE",
        type=argparse.FileType("r", encoding="UTF-8"),
        help="the IMP file to execute",
    )

    args = parser.parse_args()

    nodes = args.nodes
    depth_first = args.depth_first
    simplify = args.simplify
    assumption = parse_bexpr(args.assumption) if args.assumption is not None else None
    file_name = args.file.name
    file_contents = args.file.read()

    try:
        parsed = parse_imp(file_contents)
    except ParseCancellationException:
        print(f"Error parsing file {file_name}")
        sys.exit(DATA_FORMAT_ERROR)

    print(f"Analyzing file {file_name}\n")

    if depth_first:
        stream = find_bugs_depth_first(parsed, assumption)
    elif assumption is not None:
        stream = find_bugs_assuming(parsed, assumption)
    else:
        stream = find_bugs(parsed)

    bug_found = False
    for _ in range(nodes):
        next_result = next(stream)
        if isinstance(next_result, BugFound):
            bug_found = True
            print("BUG FOUND")
            path, store, _ = next_result.s

            if simplify:
                path = simplify_expr(path)

            print(f"  Path:          {path}")
            print(f"  Example Input: {example_input(path)}")
            print(f"  Store:         {store}")

    if not bug_found:
        print(f"No bug found at depth {nodes}.")


def example_input(path):
    solver = z3.Solver()
    solver.add(path.to_smt())
    assert solver.check() == z3.sat
    model = solver.model()

    return ", ".join(
        map(
            lambda p: f"{p[0]} := {p[1]}",
            sorted(
                {str(decl): model[decl].as_long() for decl in model.decls()}.items()
            ),
        )
    )


if __name__ == "__main__":
    main()
